require 'phi'; include Phi
require 'mci'
require 'dialogs'

# original source: Marco Cantu 1998, Mastering Delphi 4.

# constant values used by the program; if you change the
# number of items, you should resize the grid accordingly

ITEMS = 10  # items on each side of the 'square' grid
MINES = 12  # number of mines in the grid

# character codes use to describe the contents
# of the cells of the grid:
#   ?M: Mine
#   ?K: Known mine
#   ?W: Wrong mine
#   ?0..?8: Number of mines

#
# initializations
#

def init
  srand
  
  # still playing or terminated
  @playing = true
  
  # temporary bitmap
  @bmp = Bitmap.new
  
  # code of the temporary bitmap
  @last_bmp = ?_
  
  # numer of shots
  @shots = 0
  
  # mines really found
  @mines_found = 0
  
  # boolean array indicating visible elements
  @display = []
  
  # map with the codes for the cells (see above for the codes)
  @map = []
  
  #
  # empty the two arrays
  #
  
  (0...ITEMS).each do |i|
    @map[i] = []
    @display[i] = []
    (0...ITEMS).each do |j|
      @map[i][j] = ?_
      @display[i][j] = false
    end
  end
  
  #
  # place 'MINES' non-overlapping mines
  #
  
  mines_to_place = MINES
  while mines_to_place > 0
    x = rand(ITEMS)
    y = rand(ITEMS)
    # if there is not a mine
    if get_map(x, y) != ?M
      # add a mine
      set_map(x, y, ?M)
      mines_to_place -= 1
    end
  end
  
  #
  # compute the number of surrounding mines in
  # every location not having a mine
  #
  
  (0...ITEMS).each do |i|
    (0...ITEMS).each do |j|
      compute_mines(i, j) if get_map(i, j) != ?M
    end
  end
end

def get_display(x, y)
  @display[x][y]
end

def set_display(x, y, v)
  @display[x][y] = v
end

def get_map(x, y)
  @map[x][y]
end

def set_map(x, y, v)
  @map[x][y] = v
end

#
# compute the number of mines surrounding the given cell
#

def compute_mines(x, y)
  total = ?0
  # for every contiguos cell...
  (x-1).upto(x+1) do |col|
    (y-1).upto(y+1) do |row|
      # excluding those out of the borders...
      if (0...ITEMS) === col && (0...ITEMS) === row
        # if there is a mine, hidden or known,
        # increase the total surrounding mines
        case get_map(col, row)
        when ?M, ?K
          total += 1
        end
      end
    end
  end
  # store the total number of surrounding mines in the map
  set_map(x, y, total)
end

# display items near a visible zero-cell, using a recursive call

def flood_zeros(x, y)
  # double check that we are on a zero
  if get_map(x, y) != ?0
    return
  end
  # for every contiguos cell...
  (x-1).upto(x+1) do |col|
    (y-1).upto(y+1) do |row|
      # excluding out of borders and the item itself...
      if (0...ITEMS) === col && (0...ITEMS) === row &&
        !(col == x && row == y)
        # display the element, and if it is a zero, repeat the
        # operation; the code seems redundant but the program
        # needs to avoid infinite recursion with great care
        if get_map(col, row) == ?0 && get_display(col, row) == false
          # if the cell is still hidden and there is a zero
          # display it, then make the flood the zeros near the cell
          set_display(col, row, true)
          flood_zeros(col, row)
        else
          # if it is not a zero, display it
          set_display(col, row, true)
        end
        # compute the area of the cell, and invalidate it
        rect = @grid.cell_rect(col, row)
        Win.invalidate_rect(@grid.handle, rect, false)
      end
    end
  end
end

@form = Component._load open('mines.afm', 'rb').read
@grid = @form.draw_grid1

@grid.on_mouse_down = proc do |sender, button, shift, x, y|
  # get the current column and grid
  col, row = @grid.mouse_to_cell(x, y)
  # if game has ended, beep and ignore the action
  unless @playing
    Sys.beep
    break
  end
  case button
  when Phi::MB_LEFT
    # left mouse button click: shot
    @shots += 1
    @form.panel1.label_shots.caption = "Shots:#@shots"
    # if there is a mine, end the game, else display the cell
    case get_map(col, row)
    when ?M, ?K
      # mine found...
      Win.play_sound 'boom.wav', 0, Win::SND_ASYNC
      # end the game and redisplay the grid
      @playing = false
      @grid.repaint
      Phi.message_dlg "B O O M !\n\nYou have found a mine",
        Phi::MT_ERROR, [Phi::MB_OK]
    else
      # not a mine...
      # show location
      set_display(col, row, true)
      # if the click was on a 0, then show near elements
      if get_map(col, row) == ?0
        flood_zeros(col, row)
      end
    end
  else
    # right mouse button click: mine?
    case get_map(col, row)
    when ?M
      set_map(col, row, ?K)
      set_display(col, row, true)
      @mines_found += 1
      # if all mines have been found, the game ends
      if @mines_found == MINES
        Phi.message_dlg "You have won. Congratulations!",
          Phi::MT_INFORMATION, [Phi::MB_OK]
        @playing = false
        @grid.repaint
      end
    when ?K
      # if there was a known mine, the 'hidden' mine
      # is restored, and the points decremented
      set_map(col, row, ?M)
      set_display(col, row, false)
      @mines_found -= 1
    when ?0..?8
      # if there was a number, set W, wrong mine
      set_map(col, row, ?W)
      set_display(col, row, true)
    when ?W
      # if there was a wrong mine, restore the
      # number computing it again
      compute_mines(col, row)
      set_display(col, row, false)
    end
  end
  # redraw the cell of the grid
  rect = @grid.cell_rect(col, row)
  Win.invalidate_rect(@grid.handle, rect, false)
end

@grid.on_draw_cell = proc do |sender, col, row, rect, state|
  # extract the code and check its value
  code = get_map(col, row)
  # if the cell is visible
  if get_display(col, row)
    # if the code corresponds to that of the 'cached' bitmap,
    # use it, else load the new bitmap
    if code != @last_bmp
      @bmp.load code.chr+'.bmp'
      @last_bmp = code
    end
  else
    # the cell is not visible: show the default bitmap,
    # using the cache mechanism again
    if @last_bmp != ?U
      @bmp.load 'undef.bmp'
      @last_bmp = ?U
    end
  end
  # if the game is done, show the mines that were not found
  # using the cache again
  if !@playing && code == ?M
    if code != @last_bmp
      @bmp.load code.chr+'.bmp'
      @last_bmp = code
    end
  end
  @grid.canvas.draw rect.left, rect.top, @bmp
end

@form.menu.items.help1.about1.on_click = proc do |sender|
  Phi::message_dlg "Mastering Delphi Mines",
    Phi::MT_INFORMATION, [Phi::MB_OK]
end

@form.menu.items.file1.new_game1.on_click = proc do |sender|
  # reinitialize and repaint
  init
  @grid.repaint
end

@form.menu.items.file1.exit1.on_click = proc do |sender|
  @form.close
end

init
@form.show
Phi.mainloop
